/*
** 2022 November 26
**
** The author disclaims copyright to this source code.  In place of
** a legal notice, here is a blessing:
**
**    Hopefully, our data can help you create more value.
**
*/

#include <pcl/point_types.h>
#include <pcl/point_cloud.h>
#include <filesystem>
#include <atomic>
#include <rosbag/bag.h>
#include <ros/ros.h>
#include <jpeglib.h>
#include <pcl/io/pcd_io.h>

#include <pcl_conversions/pcl_conversions.h>

#include <sensor_msgs/Image.h>
#include <sensor_msgs/PointCloud2.h>
#include <sensor_msgs/Imu.h>
#include <nav_msgs/Odometry.h>
#include <sensor_msgs/CompressedImage.h>

#include <signal.h>

enum class jord_folder_type
{
    lidar,
    camera,
    none,
    miscellaneous
};

jord_folder_type check_for_folder_info(const std::filesystem::path &folder)
{
    // If the result is none，So it's an empty folder or a nonexistent folder
    jord_folder_type type{jord_folder_type::none};

    if (!std::filesystem::exists(folder)) // Folder does not exist
    {
        return type;
    }
    for (auto &p : std::filesystem::directory_iterator(folder))
    {
        if (p.is_directory())
        {
            // Contains folders?
            // There are currently no multi-level folders in the JORD dataset
            // Just ignore the folder
            continue;
        }
        if (p.path().extension() == ".pcd")
        {
            // This folder may contain the point cloud data, because it contains the pcd file
            if (type == jord_folder_type::none || type == jord_folder_type::lidar)
            {
                type = jord_folder_type::lidar;
            }
            else
            {
                // This folder seems to have everything. Isn't that normal
                type = jord_folder_type::miscellaneous;
                break;
            }
        }
        else if (p.path().extension() == ".jpg")
        {
            // This folder may contain camera data because it contains jpeg files
            if (type == jord_folder_type::none || type == jord_folder_type::camera)
            {
                type = jord_folder_type::camera;
            }
            else
            {
                // This folder seems to have everything. Isn't that normal
                type = jord_folder_type::miscellaneous;
                break;
            }
        }
        else
        {
            //There are file types in this folder that I don't recognize, so maybe it's miscellaneous?
            type = jord_folder_type::miscellaneous;
            break;
        }
    }

    return type;
}

// The JORD data set currently includes:
// npos220s data(odometery + Imu)
// Velodyne-VLP16 LiDAR data
// USB Camera data
enum class jord_object_type
{
    npos220s, 
    lidar, 
    camera,
    none // Unknown or unknown files
};

struct jord_object
{
    jord_object_type type;

    // Path to the corresponding text file
    // In the case of radar or camera, this field records the corresponding timestamp file
    // If it is npos220s data,then this field is the file corresponding to the data
    std::filesystem::path text_filepath; 

    // Folder of data
    // If it is camera or radar data, then this folder contains data for each frame
    // If it is npos220s data，then this field is null
    std::filesystem::path folder_item;
};

struct jord_root_info
{
    // All data items contained in the root path
    std::vector<jord_object> objects;

    // Unknown files or folders
    std::vector<std::filesystem::path> miscs;
};

std::filesystem::path check_timestamp_for(std::filesystem::path path)
{
    // The file name of the time information should be folder name + _timestamp.txt    
    path.replace_filename(path.filename().string() + "_timestamp.txt");
    if (std::filesystem::exists(path))
    {
        return path;
    }
    return std::filesystem::path();
}

void fill_jord_folder(jord_root_info *root, const std::filesystem::path &p)
{
    auto fi_type = check_for_folder_info(p);
    if (fi_type == jord_folder_type::miscellaneous)
    {
        root->miscs.push_back(p);
    }
    else if (fi_type == jord_folder_type::lidar)
    {
        // If this is a folder of LiDAR data (containing pcd) then there should be a corresponding timestamp file to record the time information
        auto timestamp_path = check_timestamp_for(p);
        if (!timestamp_path.empty())
        {
            root->objects.push_back({jord_object_type::lidar, timestamp_path, p});
        }
        else
        {
            printf("\033[91m;No timestamp file found for %s\033[0m;\n", p.c_str());
            root->miscs.push_back(p);
        }
    }
    else if (fi_type == jord_folder_type::camera)
    {
        // If this is a folder of camera data (containing jpg), there should also be a corresponding timestamp file to record the time information
        auto timestamp_path = check_timestamp_for(p);
        if (!timestamp_path.empty())
        {
            root->objects.push_back({jord_object_type::camera, timestamp_path, p});
        }
        else
        {
            printf("\033[91m;No timestamp file found for %s\033[0m;\n", p.c_str());
            root->miscs.push_back(p);
        }
    }
    else
    {
        printf("\033[91m;No files found for %s\033[0m;\n", p.c_str());
        root->miscs.push_back(p);
    }
}

bool jord_is_npos220s(const std::filesystem::path &p)
{
    char check_buff[16];
    std::ifstream file(p);
    file.read(check_buff, 16);
    if (file.gcount() != 16)
    {
        return false;
    }
    // JORD's npos220s data files must always start with #Odometry or #Imu to log data
    if (strncmp(check_buff, "#Odometry ", 10) == 0)
    {
        return true;
    }
    if (strncmp(check_buff, "#Imu ", 5) == 0)
    {
        return true;
    }
    return false;
}

jord_root_info check_jord_path(const std::filesystem::path &root)
{
    jord_root_info info;
    std::set<std::filesystem::path> checked_timestamp;
   
    for (auto &p : std::filesystem::directory_iterator(root))
    {
        // Filter out . , .. and any hidden files or folders
        if (p.path().filename().c_str()[0] == '.')
        {
            continue;
        }
        if (p.is_directory())
        {
            fill_jord_folder(&info, p.path());
        }
    }

    // collect all checked timestamp
    for (auto &o : info.objects)
    {
        // Record all detected timestamp files so that you can skip them later in the process
        checked_timestamp.insert(o.text_filepath);
    }

    for (auto &p : std::filesystem::directory_iterator(root))
    {
        // Filter out . , .. and any hidden files or folders
        if (p.path().filename().c_str()[0] == '.' || p.is_directory())
        {
            continue;
        }
        if(checked_timestamp.find(p.path()) != checked_timestamp.end()) 
            continue;
        
        if (p.path().extension() == ".txt")
        {
            if (jord_is_npos220s(p.path()))
            {
                /// npos220s data is a separate txt file
                info.objects.push_back({jord_object_type::npos220s, p.path(), std::filesystem::path()});
                continue;
            }
        }

        // Could it be a miscellaneous file?
        info.miscs.push_back(p.path());
    }  
    return info;
}

void print_jord_root_info(const jord_root_info &info)
{
    printf("Objects:\n");
    for (auto &o : info.objects)
    {
        printf("\t%s\t%s\t%s\n", o.text_filepath.c_str(), o.folder_item.c_str(), o.type == jord_object_type::npos220s ? "npos220s" : o.type == jord_object_type::lidar ? "lidar" : o.type == jord_object_type::camera ? "camera" : "none");
    }
    printf("Miscs:\n");
    for (auto &m : info.miscs)
    {
        printf("\t%s\n", m.c_str());
    }
}

static auto jord_read_timestamp(const std::filesystem::path &p)
{
    std::vector<std::pair<ros::Time, ros::Time>> timestamps;
    std::ifstream file(p);
    std::string line;
    try
    {
        // Each row is a piece of data, corresponding to a frame's timestamp
        while (std::getline(file, line))
        {
            ros::Time t;
            size_t end_pos = 0;
            t.fromSec(std::stod(line, &end_pos));
            // If there is still a double behind, then it means that we have read the hardware time successfully, and the rest is the arrival time.
            if(line.find_first_of("0123456789", end_pos) != std::string::npos){
                ros::Time t2;
                t2.fromSec(std::strtod(line.data() + end_pos, nullptr));
                timestamps.push_back({t, t2});
            } else {
                timestamps.push_back({t, t});
            }
        }
    }
    catch (std::exception &e)
    {
        printf("\033[92m;Error reading timestamp file %s\033[0m;\n", p.c_str());
        timestamps.clear();
    }
    return timestamps;
}

struct jord_npos220s_value
{
    // timestamp
    ros::Time timestamp;
    // true corresponds to IMU data
    // false corresponds to the odometery data
    bool is_imu;
    union
    {
        struct
        {
            // Three axis acceleration
            float ax, ay, az;
            // Triaxial angular velocity
            float gx, gy, gz;
        } imu;

        struct
        {
            // Odometery position data
            float x, y, z;
            // Quaternion data of odometery attitude
            float qx, qy, qz, qw;
            // Standard deviation of the odometer position
            float std_x, std_y, std_z;
        } odom;
    };
};

/*
    JORD's IMU data will start with #Imu, first with a Unix Timestamp accurate to milliseconds, then with three floating-point values representing triaxial angular velocity, and finally with three floating-point values holding the triaxial acceleration.
    JORD does not save GPS data directly, but rather in the form of an odometer in the NED coordinate system with reference to the first frame of each packet.
    JORD's Odometry data is stored in #Odometry, which is the same as the IMU: the first number is a Unix Timestamp accurate to the millisecond, but the relative position (x, y, z) is followed by a quaternion for the pose (based on the NED coordinate system).
         The last three numbers are the standard deviation of the error relative to the position.
*/
std::vector<jord_npos220s_value> jord_read_npos220s(const std::filesystem::path &p)
{
    std::vector<jord_npos220s_value> values;
    FILE *file = fopen(p.c_str(), "r");
    if (file == nullptr)
    {
        printf("\033[92m;Error opening npos220s file %s\033[0m;\n", p.c_str());
        return values;
    }
    char line[1024];//Theoretically, a row of data should not exceed 1024 bytes, and if it does, we can infer that there is an error in the data.

    while (fgets(line, 1024, file))
    {
        jord_npos220s_value v;
        double timestamp;
        if (strncmp(line, "#Imu ", 5) == 0)
        {
            if (sscanf(line, "#Imu %lf %f %f %f %f %f %f", &timestamp, &v.imu.gx, &v.imu.gy, &v.imu.gz, &v.imu.ax, &v.imu.ay, &v.imu.az) != 7)
            {
                printf("\033[92m;Error reading npos220s #Imu %s(%s)\033[0m;\n", p.c_str(), line);
                values.clear();
                break;
            }
            v.is_imu = true;
        }
        else if (strncmp(line, "#Odometry ", 10) == 0)
        {
            if (sscanf(line, "#Odometry %lf %f %f %f %f %f %f %f %f %f %f", &timestamp, &v.odom.x, &v.odom.y, &v.odom.z, &v.odom.qx, &v.odom.qy, &v.odom.qz, &v.odom.qw, &v.odom.std_x, &v.odom.std_y, &v.odom.std_z) != 11)
            {
                printf("\033[92m;Error reading npos220s #Odometry %s(%s)\033[0m;\n", p.c_str(), line);
                values.clear();
                break;
            }
            v.is_imu = false;
        }
        else
        {
            continue;
        }
        v.timestamp.fromSec(timestamp);
        values.push_back(v);
    }

    fclose(file);
    return values;
}

/*
the folder of JORD contains a lot of frame data, increasing from 0, so we need to sort it to get the correct frame order
*/
auto sort_numberic_filename(std::vector<std::filesystem::path> &paths)
{
    // In order to ensure the correctness of our subsequent stol operation, we need to exclude data that does not start with the file name
    auto end = std::remove_if(paths.begin(), paths.end(), [](const std::filesystem::path &p)
                              {
        auto fs = p.filename().string();
        auto n = fs.find_first_not_of("0123456789");
        // Illegal file names include three possibilities
        // 1.All numbers (no extension) or empty string
        // 2. The first one of the file name. Contains non-numeric characters before (such as 123a.pcd)
        // 3. Does not start with a number (contains only the extension, such as .pcd)
        // Here we have not judged whether there are file names like 123.abc.pcd, tentatively they are legal.
        if(n == std::string::npos || fs[n] != '.' || n == 0)
        {
            return true;
        }
        return false; });

        // This sorting performance may not be very high, but the problem is not big
    std::sort(paths.begin(), end, [](const std::filesystem::path &a, const std::filesystem::path &b)
              {
        auto a_str = a.filename().string();
        auto b_str = b.filename().string();
        auto a_num = std::stol(a_str.substr(0, a_str.find('.')));
        auto b_num = std::stol(b_str.substr(0, b_str.find('.')));
        return a_num < b_num; });

    return end;
}

/*
Data source, data source including npos220s data source, radar data source, camera data source
*/
struct jord_source
{
    virtual ~jord_source() {}

    // Publishing ros Messages
    virtual void publish() = 0; 
    
    // Is there any more news
    virtual bool more() const = 0 ; 
    
    // Write the data to the bag
    virtual void write_bag(rosbag::Bag &bag) const = 0;

    //next
    virtual void next() = 0;
    
    //Time of the current frame
    virtual ros::Time time() const = 0;
};

// Data sources from folders, including camera and LiDAR data sources
struct jord_folder_source : jord_source
{
    std::vector<std::filesystem::path> paths;//  All files
    std::vector<std::pair<ros::Time, ros::Time>> timestamps; // 
    size_t next_frame;

    // tag contains the name of the item that will be used to generate the topic
    // This field is derived from the folder name.
    std::string tag_name;

    bool initialize(ros::NodeHandle *nh, std::filesystem::path &folder, const std::filesystem::path &timestamp_file)
    {
        // Get information about all files in the folder
        for (auto &p : std::filesystem::directory_iterator(folder))
        {
            // Filter out all hidden files and paths
            if (p.path().filename().c_str()[0] == '.' || p.is_directory())
            {
                continue;
            }
            paths.push_back(p.path());
        }
        auto end = sort_numberic_filename(paths);
        // Including a file that is not named with a number means that the file data is corrupted.
        if (end != paths.end())
        {
            printf("\033[91mError: folder %s contains non-numeric filename\033[0m\n", folder.c_str());
            return false;
        }
        timestamps = jord_read_timestamp(timestamp_file);

        // Each frame in the timestamp data corresponds to one frame of data.
        // If the data frame is incomplete, that means the data may be missing
        if (paths.size() != timestamps.size())
        {
            printf("\033[91m(%s)Error: count %lu != timestamp count %lu\033[0m;\n", folder.c_str(), paths.size(), timestamps.size());
            return false;
        }
        next_frame = 0;
        
        tag_name = folder.filename().string();
        if (tag_name.empty())
        {
            printf("\033[91mError: folder %s is empty\033[0m;\n", folder.c_str());
            return false;
        }

        //The topic consists of /jord/folder_name (such as /jord/camera1)
        tag_name.insert(0, "/jord/");
        return true;
    }

    void publish() override
    {
        if (next_frame >= paths.size())
        {
            return;
        }

        //Publishes the current frame data
        do_publish(paths[next_frame], timestamps[next_frame].second);
    }

    bool more() const override
    {
        return next_frame < paths.size() - 1;
    }

    void write_bag(rosbag::Bag &bag) const override
    {
        if (next_frame >= paths.size())
        {
            return;
        }
        do_write(bag, paths[next_frame], timestamps[next_frame].first, timestamps[next_frame].second);
    }

    void next() override
    {
        next_frame++;
    }

    ros::Time time() const override
    {
        return timestamps[next_frame].second;
    }

    //Data corresponding to p is published
    virtual void do_publish(const std::filesystem::path &p, const ros::Time &timestamp) = 0;

    //Write the data into the bag
    virtual void do_write(rosbag::Bag &bag, const std::filesystem::path &p, const ros::Time &timestamp, const ros::Time& arrival_time) const = 0;
};

// Decode the jpeg file to sensor_msgs::Image
// Attach the URL of libjpeg
bool decode_jpeg_io(FILE* fp, sensor_msgs::Image *image)
{
    struct jpeg_decompress_struct cinfo;
    struct jpeg_error_mgr jerr;
    cinfo.err = jpeg_std_error(&jerr);
    jpeg_create_decompress(&cinfo);
    jpeg_stdio_src(&cinfo, fp);
    if (jpeg_read_header(&cinfo, TRUE) != JPEG_HEADER_OK)
    {
        printf("\033[31mFailed to read jpeg header\033[0m\n");
        return false;
    }
    if (jpeg_start_decompress(&cinfo) != TRUE)
    {
        printf("\033[31mFailed to start decompress\033[0m\n");
        return false;
    }
    image->width = cinfo.output_width;
    image->height = cinfo.output_height;
    image->encoding = "rgb8";
    image->step = image->width * 3;
    image->data.resize(image->width * image->height * 3);
    while (cinfo.output_scanline < cinfo.output_height)
    {
        auto row = cinfo.output_scanline;
        unsigned char *ptr = image->data.data() + row * image->step;
        jpeg_read_scanlines(&cinfo, (JSAMPARRAY)&ptr, 1);
    }
    jpeg_finish_decompress(&cinfo);
    jpeg_destroy_decompress(&cinfo);
    return true;
}

bool decode_jpeg(const std::filesystem::path &p, sensor_msgs::Image *image) {
    FILE* fp = fopen(p.c_str(), "rb");
    if(fp == nullptr) {
        printf("\033[92m;Error: cannot open file %s\033[0m;\n", p.c_str());
        return false;
    }

    auto result = decode_jpeg_io(fp, image);
    fclose(fp);

    return result;
}
// Read the contents of a file
std::string read_full_file(const std::filesystem::path &p)
{
    FILE *file = fopen(p.c_str(), "rb");
    if (!file)
    {
        printf("\033[92m;Error: cannot open file %s\033[0m;\n", p.c_str());
        return "";
    }
    fseek(file, 0, SEEK_END);
    auto size = ftell(file);
    fseek(file, 0, SEEK_SET);
    std::string buffer;
    buffer.resize(size);
    auto readed = fread(buffer.data(), 1, size, file);
    fclose(file);
    if (readed != size)
    {
        printf("\033[92m;Error: readed %lu != size %lu\033[0m;\n", readed, size);
        return "";
    }
    return buffer;
}

// JORD Camera data source
struct jord_camera_source : jord_folder_source
{
    ros::Publisher pub;

    bool write_compress = true;

    bool initialize(ros::NodeHandle *nh, std::filesystem::path &folder, const std::filesystem::path &timestamp_file)
    {
        if (!jord_folder_source::initialize(nh, folder, timestamp_file))
        {
            return false;
        }

        if (nh != nullptr)
        {
            // Only the decoded data can be published.
            pub = nh->advertise<sensor_msgs::Image>(tag_name, 1);
        }
        return true;
    }

    void do_publish(const std::filesystem::path &p, const ros::Time &t) override
    {
        sensor_msgs::Image image;
        image.header.stamp = t;
        image.header.frame_id = "camera";
        if (!decode_jpeg(p, &image))
        {
            printf("\033[31mFailed to decode jpeg\033[0m\n");
            return;
        }
        pub.publish(image);
    }

    void do_write(rosbag::Bag &bag, const std::filesystem::path &p, const ros::Time &t, const ros::Time& arrival_time) const override
    {
        if (write_compress)
        {
            auto buffer = read_full_file(p);
            if(buffer.empty())
                return;
            sensor_msgs::CompressedImage image;
            image.header.stamp = t;
            image.header.frame_id = "camera";
            image.format = "jpeg";
            image.data = std::vector<uint8_t>(buffer.begin(), buffer.end());
            bag.write(tag_name + "/compressed", arrival_time, image);
        }
        else
        {
            sensor_msgs::Image image;
            image.header.stamp = t;
            image.header.frame_id = "camera";
            if (!decode_jpeg(p, &image))
            {
                printf("\033[31mFailed to decode jpeg\033[0m");
                return;
            }
            bag.write(tag_name, arrival_time, image);
        }
    }
};

struct jord_lidar_source : jord_folder_source
{
    ros::Publisher pub;

    bool initialize(ros::NodeHandle *nh, std::filesystem::path &folder, const std::filesystem::path &timestamp_file)
    {
        if (!jord_folder_source::initialize(nh, folder, timestamp_file))
        {
            return false;
        }

        if (nh != nullptr)
        {
            pub = nh->advertise<sensor_msgs::PointCloud2>(tag_name, 1);
        }
        return true;
    }

    static sensor_msgs::PointCloud2 read_pcd(const std::filesystem::path &p, const ros::Time &t)
    {
        // Generate by GitHub Copilot
        sensor_msgs::PointCloud2 cloud;
        pcl::PCLPointCloud2 pcl_pc2;
        pcl::io::loadPCDFile(p.string(), pcl_pc2);
        pcl_conversions::fromPCL(pcl_pc2, cloud);
        cloud.header.stamp = t;
        cloud.header.frame_id = "velodyne";
        return std::move(cloud);
    }

    void do_publish(const std::filesystem::path &p, const ros::Time &t) override
    {
        auto cloud = read_pcd(p, t);
        pub.publish(cloud);
    }

    void do_write(rosbag::Bag &bag, const std::filesystem::path &p, const ros::Time &t, const ros::Time& arrival_time) const override
    {
        auto cloud = read_pcd(p, t);
        bag.write(tag_name, arrival_time, cloud);
    }
};

// JORD npos220s data source
struct jord_npos220s_source : jord_source
{
    ros::Publisher imu_pub, odom_pub; 
    std::vector<jord_npos220s_value> values;
    size_t next_frame;

    // At present, our data set contains only one npos220s, so this topic can be fixed to these two items
    static constexpr const char* imu_topic_name = "/jord/npos220s/imu";
    static constexpr const char* odom_topic_name = "/jord/npos220s/odometry";

    bool initialize(ros::NodeHandle *nh, const std::filesystem::path &text_file)
    {
        next_frame = 0;
        // Since the npos220s usually has a small amount of data, we can pre-load all the data into memory for use.
        values = jord_read_npos220s(text_file);
        if (values.empty())
        {
            return false;
        }

        if (nh != nullptr)
        {
            imu_pub = nh->advertise<sensor_msgs::Imu>(imu_topic_name, 1);
            odom_pub = nh->advertise<nav_msgs::Odometry>(odom_topic_name, 1);
        }
        return true;
    }

    static sensor_msgs::Imu get_imu(const jord_npos220s_value &v)
    {
        sensor_msgs::Imu imu;
        imu.header.stamp = v.timestamp;
        imu.header.frame_id = "base_link";
        imu.linear_acceleration.x = v.imu.ax;
        imu.linear_acceleration.y = v.imu.ay;
        imu.linear_acceleration.z = v.imu.az;
        imu.angular_velocity.x = v.imu.gx;
        imu.angular_velocity.y = v.imu.gy;
        imu.angular_velocity.z = v.imu.gz;
        return imu;
    }

    static nav_msgs::Odometry get_odom(const jord_npos220s_value &v)
    {
        nav_msgs::Odometry odom;
        odom.header.stamp = v.timestamp;
        odom.header.frame_id = "odom";
        odom.child_frame_id = "base_link";
        odom.pose.pose.position.x = v.odom.x;
        odom.pose.pose.position.y = v.odom.y;
        odom.pose.pose.position.z = v.odom.z;
        odom.pose.pose.orientation.x = v.odom.qx;
        odom.pose.pose.orientation.y = v.odom.qy;
        odom.pose.pose.orientation.z = v.odom.qz;
        odom.pose.pose.orientation.w = v.odom.qw;
        // The rest of this covariance matrix is going to be initialized to 0
        odom.pose.covariance[0] = v.odom.std_x * v.odom.std_x;
        odom.pose.covariance[7] = v.odom.std_y * v.odom.std_y;
        odom.pose.covariance[14] = v.odom.std_z * v.odom.std_z;
        return odom;
    }
    void publish() override
    {
        if (next_frame >= values.size())
        {
            return;
        }

        auto &v = values[next_frame];
        if (v.is_imu)
        {
            imu_pub.publish(get_imu(v));
        }
        else
        {
            odom_pub.publish(get_odom(v));
        }
    }

    bool more() const override
    {
        return next_frame < values.size() - 1;
    }

    void write_bag(rosbag::Bag &bag) const
    {
        if (next_frame >= values.size())
        {
            return;
        }

        auto &v = values[next_frame];
        if (v.is_imu)
        {
            auto imu = get_imu(v);
            bag.write(imu_topic_name, v.timestamp, imu);
        }
        else
        {
            auto odom = get_odom(v);
            bag.write(odom_topic_name, v.timestamp, odom);
        }
    }

    void next() override {
        next_frame++;
    }

    ros::Time time() const override {
        return values[next_frame].timestamp;
    }
};

using pq_item = std::pair<ros::Time, size_t>;

struct pq_item_cmp
{
    bool operator()(const pq_item &a, const pq_item &b) const
    {
        return a.first > b.first;
    }
};

// Collate and sort all the data sources, one by one
struct jord_object_reader
{
    std::vector<std::unique_ptr<jord_source>> sources;
    ros::NodeHandle *nh = nullptr;

    // There is only one entry per data source in the priority queue
    std::priority_queue<pq_item, std::vector<pq_item>, pq_item_cmp> queue;
};

bool jor_append_source(jord_object_reader* jor, size_t source_id) {
    // Ensure that the id is valid
    if (source_id >= jor->sources.size()) {
        return false;
    }
    auto &s = jor->sources[source_id];
    // There is no more data in this data source
    if (!s->more()) {
        return false;
    }

    jor->queue.push({s->time(), source_id});
    return true;
}

bool jor_initialize(jord_object_reader *jor, ros::NodeHandle *nh, const std::filesystem::path &folder, bool image_compress)
{
    jor->nh = nh;
    // Gets project information under the root path
    auto info = check_jord_path(folder);
    // Output this information so that we can verify that everything is the same as what we want
    print_jord_root_info(info);
    for (auto var : info.objects)
    {
        switch (var.type)
        {
        case jord_object_type::camera:
        {
            auto source = std::make_unique<jord_camera_source>();
            if (!source->initialize(nh, var.folder_item, var.text_filepath))
            {
                return false;
            }
            
            // Add the camera source to the data source list
            // write_compress is used to determine whether to compress the image when writing to the bag file
            source->write_compress = image_compress;
            jor->sources.push_back(std::move(source));
            jor_append_source(jor, jor->sources.size() - 1);
            break;
        }
        case jord_object_type::lidar:
        {
            auto source = std::make_unique<jord_lidar_source>();
            if (!source->initialize(nh, var.folder_item, var.text_filepath))
            {
                return false;
            }
            jor->sources.push_back(std::move(source));
            jor_append_source(jor, jor->sources.size() - 1);
            break;
        }
        case jord_object_type::npos220s:
        {
            auto source = std::make_unique<jord_npos220s_source>();
            if (!source->initialize(nh, var.text_filepath))
            {
                return false;
            }
            jor->sources.push_back(std::move(source));
            jor_append_source(jor, jor->sources.size() - 1);
            break;
        }
        }
    }

    return true;
}

bool jor_more(jord_object_reader *jor)
{
    return !jor->queue.empty();
}

void jor_play(jord_object_reader *jor)
{
    auto top = jor->queue.top();
    jor->queue.pop();
    auto &s = jor->sources[top.second];
    s->publish();
    s->next();
    jor_append_source(jor, top.second);
}

void jor_write_bag(jord_object_reader *jor, rosbag::Bag &bag)
{
    auto top = jor->queue.top();
    jor->queue.pop();
    auto &s = jor->sources[top.second];
    s->write_bag(bag);
    s->next();
    jor_append_source(jor, top.second);

}

struct jord_play_config {
    std::filesystem::path folder;
    float speed = 1.0f;
};

bool parse_jord_play_argv(int argc, const char* const * argv, jord_play_config* cfg) {
    for(int i = 2; i < argc; i++) {
        if(argv[i][0] != '-') {
            if(cfg->folder.empty())
                cfg->folder = argv[i];
            else {
                printf("Unexpected argument: %s\n", argv[i]);
                return false;
            }
            continue;
        }
        if (strcmp(argv[i], "--speed") == 0 || strcmp(argv[i], "-s") == 0) {
            if (i + 1 >= argc) {
                printf("Missing speed value after %s\n", argv[i]);
                return false;
            }
            try {
                cfg->speed = std::stof(argv[++i]);
            } catch (...) {
                printf("Invalid speed value: %s\n", argv[i]);
                return false;
            }
            continue;
        } else {
            printf("Unknown argument: %s\n", argv[i]);
            return false;
        }
    }
    if(cfg->folder.empty()) {
        printf("Missing folder argument\n");
        return false;
    }
    return true;
}

struct jord_play_status {
    ros::Time start_play_time;
    ros::Time first_frame_time;
};

void jps_wait_until(jord_play_status* jps, ros::Time t, float speed) {
    if(jps->first_frame_time.is_zero()) {
        jps->first_frame_time = t;
        jps->start_play_time = ros::Time::now();
        return;
    }
    auto now = ros::Time::now();
    // The elapsed time between the start of the play and the present
    auto elapsed = now - jps->start_play_time;
    // The length of the target time relative to the start time
    auto expected = (t - jps->first_frame_time)*  (1.0f / speed);
    if (expected > elapsed) {
        // Wait until finished.
        ros::Duration(expected - elapsed).sleep();
    }
}

// The implementation of jord play
int jord_play(int argc, const char* const * argv) {
    // 1. Read the parameters and initialize the configuration
    jord_play_config cfg;
    if(!parse_jord_play_argv(argc, argv, &cfg)) {
        return 1;
    }
    // 2. Initializing ros
    ros::init(argc, const_cast<char**>(argv), "jord_play");
    ros::NodeHandle nh;

    // 3. Open the JORD dataset
    jord_object_reader jor;
    if(!jor_initialize(&jor, &nh, cfg.folder, false)) {
        printf("Failed to initialize jord object reader\n");
        return 1;
    }

    // 4. Read each frame, wait until the appropriate time and play.
    jord_play_status jps;
    while(jor_more(&jor)) {
        if(!ros::ok()) break;
        jps_wait_until(&jps, jor.queue.top().first, cfg.speed);
        jor_play(&jor);
    }

    return 0;
}

struct jord_write_bag_config {
    std::filesystem::path folder;
    std::filesystem::path bag_path;

    // If true, the contents of the image are compressed CompressedImage 
    //Otherwise, the Image is uncompressed
    bool compressed = true;
};

bool parse_jord_write_bag_argv(int argc, const char* const * argv, jord_write_bag_config* cfg) {
    for(int i = 2; i < argc; i++) {
        if(argv[i][0] != '-') {
            // 2 POSITION ARGs
            if(cfg->folder.empty())
                cfg->folder = argv[i];
            else if(cfg->bag_path.empty())
                cfg->bag_path = argv[i];
            else {
                printf("Unexpected argument: %s\n", argv[i]);
                return false;
            }
            continue;
        }
        if (strcmp(argv[i], "--uncompressed") == 0 || strcmp(argv[i], "-u") == 0) {
            cfg->compressed = false;
            continue;
        } else {
            printf("Unknown argument: %s\n", argv[i]);
            return false;
        }
    }
    if(cfg->folder.empty()) {
        printf("Missing folder argument\n");
        return false;
    }

    if(cfg->bag_path.empty()) {
        printf("Missing bag path argument\n");
        return false;
    }
    return true;
}

// This variable is used to record whether the user wants to stop writing bags. When Ctrl+C is pressed, this flag is set to true.
static std::atomic_bool stop_flag;

static void SIGINT_handler(int sig) {
    stop_flag = true;
}

int jord_write_bag(int argc, const char* const * argv) {
    // 1. Read the parameters and initialize the configuration
    jord_write_bag_config cfg;
    if(!parse_jord_write_bag_argv(argc, argv, &cfg)) {
        return 1;
    }

    // 2. Open the JORD dataset
    jord_object_reader jor;
    if(!jor_initialize(&jor, nullptr, cfg.folder, cfg.compressed)) {
        printf("Failed to initialize jord object reader\n");
        return 1;
    }

    // 3.Open the bag to write data
    rosbag::Bag bag(cfg.bag_path, rosbag::bagmode::Write);
    if(!bag.isOpen()) {
        printf("Failed to Open Bag for write\n");
        return 1;
    }
    
    // 4. Setup the SIGINT handler for exit.
    stop_flag = false;
    signal(SIGINT, SIGINT_handler);

    // 5. Constantly reading data and writing packets
    while(jor_more(&jor)) {
        if(stop_flag.load()) 
            break;
        jor_write_bag(&jor, bag);
    }

    // 5. Finish writing
    bag.close();
    return 0;
}

void print_usage() {
    printf("Usage: jord <command> [args]\r\n");
    printf("Commands:\r\n");
    printf("  play <folder> [--speed <speed>]\r\n");
    printf("  write-bag <folder> <bag_path> [--uncompressed]\r\n");
}

int main(int argc, const char* const * argv) {
    if(argc < 2) {
        print_usage();
        return 1;
    }
    if(strcmp(argv[1], "play") == 0) {
        return jord_play(argc, argv);
    } else if(strcmp(argv[1], "write-bag") == 0) {
        return jord_write_bag(argc, argv);
    } else {
        printf("Unknown command: %s\r", argv[1]);
        print_usage();
        return 1;
    }
}
